#define Uses_IdString
#define Uses_RefString
#define Uses_RefStrCollection
#define Uses_TVMemMgr
#define Uses_TStreamableClass
#include "LibStart.h"
#include <tv.h>
#include "LibEnd.h"
#include <string.h>

REGISTER_STREAMABLE(RefStrCollection)
STREAMABLE_BUILD(RefStrCollection)

RefStrCollection::RefStrCollection(const IdString* ids) :
	TNSCollection(0,0),
	bufSize(0),
	base(0)
{
	ushort n;
	int len;
	RefString* rs;
	long bsize = 0L;

	shouldDelete = False;
	if(ids != 0)
		for(n = 0; ids[n].text != 0; n++)
			bsize += sizeof(ushort)+strlen(ids[n].text)+1;
	if(bsize > MaxRefStrBufSize)
		return;

	bufSize = bsize;
	// Not a bug - this ctor is meant to only be used in a resource-creation
	//	program - abort() for lack of memory is OK.
	if(bufSize)
		{
		base = new char [bufSize];
		setLimit(n);
		char* buf = base;
		for(n = 0; ids[n].text != 0; n++)
			{
			rs = (RefString*) buf;
			rs->id = ids[n].id;
			strcpy(rs->text, ids[n].text);
			len = strlen(ids[n].text);
			insert(buf);
			buf += len+sizeof(ushort)+1;
			}
		}
}

void RefStrCollection::shutDown()
{
	count = 0;
	delete [] base;
	TNSCollection::shutDown();
}

const char* RefStrCollection::get(ushort id)
{
	ccIndex idx;
	if(search(&id, idx))
		{
		RefString* rs = (RefString*) at(idx);
		return rs->text;
		}
	else
		return 0;
}

void RefStrCollection::freeItem(void*)
{}

int RefStrCollection::compare(void* key1, void* key2)
{
	RefString* rs1 = (RefString*) key1;
	RefString* rs2 = (RefString*) key2;
	if(rs1->id < rs2->id)
		return -1;
	else if(rs1->id > rs2->id)
		return 1;
	else
		return 0;
}

RefStrCollection::RefStrCollection(StreamableInit)
{}

void RefStrCollection::write(opstream& os)
{
	os << count << limit << delta << bufSize;
	os.writeBytes(base, bufSize);
	for(ccIndex n = 0; n < count; n++)
		os << ushort( (char*)(items[n]) - base );
}

void* RefStrCollection::read(ipstream& is)
{
	ushort idx;
	ccIndex n;

	is >> count >> idx >> delta >> bufSize;

//	base = new char [bufSize];

//	This was a bug! I was hoping someone would notice and ask about it -
//	since they didn't, I'll "repair" it.
//  The problem: bufSize might be a large amount - up to almost 64KB. If
//	this "major consumer" cannot allocate, TV's ::new() will abort() the
//	program - we *don't* want this! So we will first resize the safety pool
//	to account for its size, and just stop (fail) if we can't resize the
//	safety pool.

	// Take into account both the allocation from setLimit() and the alloc
	//	for base.
	long total = long(idx) * sizeof(void*) + 32 + bufSize + 32;
	if(total <= 65000L)
		TVMemMgr::resizeSafetyPool(total);
	if(total > 65000L || TVMemMgr::safetyPoolExhausted())
		{
		// Resize back to old size.
		TVMemMgr::resizeSafetyPool();
		// Read remaining data so as not to screw up the resource file.
		for(n = 0; n < count; n++)
			is >> idx;
		// Indicate that its empty.
		bufSize = 0;
		count = 0;
		setLimit(0);
		}

	// Things are OK - finish up.
	else
		{
		base = new char [bufSize];
		is.readBytes(base, bufSize);
		setLimit(idx);
		for(n = 0; n < count; n++)
			{
			is >> idx;
			items[n] = base + idx;
			}
		TVMemMgr::resizeSafetyPool();
		}
	return this;
}

